from vaspvis.unfold import make_kpath,removeDuplicateKpoints, find_K_from_k, save2VaspKPOINTS
from vaspvis.unfold import convert
from vaspvis.passivator_utils import _append_H, _cart2sph, _get_bot_index, _get_neighbors, _get_top_index,_sort_by_z, _sph2cart, _center_slab
from pymatgen.analysis.graphs import StructureGraph
from pymatgen.analysis.local_env import JmolNN, CrystalNN, EconNN
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.molecule_structure_comparator import CovalentRadius
from pymatgen.core.periodic_table import Element
from pymatgen.io.vasp.outputs import Eigenval, BSVasprun
from pymatgen.io.vasp.inputs import Incar, Kpoints
from pymatgen.electronic_structure.core import Spin, Orbital
from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.io.ase import AseAtomsAdaptor
from ase.build import surface
from ase.build import niggli_reduce, sort
from ase.io import read, write
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.molecule_structure_comparator import CovalentRadius
from pyprocar.utilsprocar import UtilsProcar
from pyprocar.procarparser import ProcarParser
import matplotlib.pyplot as plt
from fastdtw import fastdtw
import numpy as np
import time
import copy
import os


def group_layers(structure, atol=None):
    """
    This function will find the atom indices belonging to each unique atomic layer.

    Parameters:
        structure (pymatgen.core.structure.Structure): Slab structure
        atol (float or None): Tolarence used for grouping the layers. Useful for grouping
            layers in a structure with relaxed atomic positions.

    Returns:
        A list containing the indices of each layers.
        A list of heights of each layers in fractional coordinates.
    """
    sites = structure.sites
    zvals = np.array([site.c for site in sites])
    unique_values = np.sort(np.unique(np.round(zvals, 3)))
    diff = np.mean(np.diff(unique_values)) * 0.2

    grouped = False
    groups = []
    group_heights = []
    zvals_copy = copy.deepcopy(zvals)
    while not grouped:
        if len(zvals_copy) > 0:
            if atol is None:
                group_index = np.where(
                    np.isclose(zvals, np.min(zvals_copy), atol=diff)
                )[0]
            else:
                group_index = np.where(
                    np.isclose(zvals, np.min(zvals_copy), atol=atol)
                )[0]

            group_heights.append(np.min(zvals_copy))
            zvals_copy = np.delete(zvals_copy, np.where(
                np.isin(zvals_copy, zvals[group_index]))[0])
            groups.append(group_index)
        else:
            grouped = True

    return groups, np.array(group_heights)


def convert_slab(bulk_path, slab_path, index, output='POSCAR_unfold', generate=True, print_M=True):
    """
    This function rotates a slab structure so its transformation matrix
    (M) to the primitive bulk structure becomes an integer matrix

    Parameters:
        bulk_path (str): File path to a primitive bulk POSCAR file
        slab_path (str): File patch to a slab structure
        index (list): A three element list defining the miller index of the surface
        output (str): File name of the converted slab POSCAR

    Returns:
        Converted slab POSCAR
    """
    M = convert.convert(
        bulk=bulk_path,
        slab=slab_path,
        index=index,
        output=output,
        generate=generate,
        print_M=print_M,
    )

    return M

def generate_kpoints(M, high_symmetry_points, n, output='KPOINTS'):
    """
    This function generates a KPOINTS file for a band unfolding calculation

    Parameters:
        M (list[list]): A 3x3 transformation matrix
        high_symmetry_points (list[list]): Fractional coordinated of the high symmetry points
            in the band structure path.
        n (int): Numbering of segments between each point
        output (str): File name of the KPOINTS file

    Returns:
        KPOINTS file
    """
    kpath = make_kpath(high_symmetry_points, nseg=n)
    K_in_sup = []
    for kk in kpath:
        kg, _ = find_K_from_k(kk, M)
        K_in_sup.append(kg)
    reducedK = removeDuplicateKpoints(K_in_sup)

    save2VaspKPOINTS(reducedK, output)


class BandGap():
    """
    Determines the band gap from a band structure or density of states calculation

    Parameters:
        folder (str): Folder that contains the VASP input and outputs files
        spin (str): 'both' returns the bandgap for all spins, 'up' returns the 
            bandgap for only the spin up states and 'down' returns the bandgap
            for only the spin down states.
        soc_axis (None or str): This parameter can either take the value of None or the
            it can take the value of 'x', 'y', or 'z'. If either 'x', 'y', or 'z' are given
            then spin='up' states will be defined by positive values of this spin-component
            and spin='down' states will be defined by negative values of this spin-component.
            This will only be used for showing a pseudo-spin-polarized plot for calculations
            that have SOC enabled.
        method (int): method=0 gets the band gap by finding the values closest to
            the fermi level. method=1 gets the band gap based on the average energy
            of each band.
    """
    def __init__(self, folder, spin='both', soc_axis=None, method=0) -> None:
        self.folder = folder
        self.method = method
        self.spin = spin
        self.soc_axis = soc_axis
        self.eigenval = Eigenval(os.path.join(folder, 'EIGENVAL'))
        self.efermi = float(os.popen(f'grep E-fermi {os.path.join(folder, "OUTCAR")}').read().split()[2])

        self.incar = Incar.from_file(
            os.path.join(folder, 'INCAR')
        )
        if 'LSORBIT' in self.incar:
            if self.incar['LSORBIT']:
                self.lsorbit = True
            else:
                self.lsorbit = False
        else:
            self.lsorbit = False

        if 'ISPIN' in self.incar:
            if self.incar['ISPIN'] == 2:
                self.ispin = True
            else:
                self.ispin = False
        else:
            self.ispin = False

        if 'LHFCALC' in self.incar:
            if self.incar['LHFCALC']:
                self.hse = True
            else:
                self.hse = False
        else:
            self.hse = False

        self.pre_loaded_bands = os.path.isfile(os.path.join(folder, 'eigenvalues.npy'))

        if soc_axis is not None and self.lsorbit:
            self.pre_loaded_spin_projections = os.path.isfile(os.path.join(folder, 'spin_projections.npy'))

        self.bg, self.vbm, self.cbm = self._get_bandgap(method=self.method)


    def _load_soc_spin_projection(self):
        """
        This function loads the project weights of the orbitals in each band
        from vasprun.xml into a dictionary of the form:
        band index --> atom index --> weights of orbitals

        Returns:
            projected_dict (dict([str][int][pd.DataFrame])): Dictionary containing the projected weights of all orbitals on each atom for each band.
        """
        if not self.lsorbit:
            raise BaseException(f"You selected soc_axis='{self.soc_axis}' for a non-soc axis calculation, please set soc_axis=None")
        if self.lsorbit and self.soc_axis == 'x':
            spin = 1
        elif self.lsorbit and self.soc_axis == 'y':
            spin = 2
        elif self.lsorbit and self.soc_axis == 'z':
            spin = 3

        if not os.path.isfile(os.path.join(self.folder, 'PROCAR_repaired')):
            UtilsProcar().ProcarRepair(
                os.path.join(self.folder, 'PROCAR'),
                os.path.join(self.folder, 'PROCAR_repaired'),
            )

        if self.pre_loaded_spin_projections:
            with open(os.path.join(self.folder, 'spin_projections.npy'), 'rb') as spin_projs:
                spin_projections = np.load(spin_projs) 
        else:
            parser = ProcarParser()
            parser.readFile(os.path.join(self.folder, 'PROCAR_repaired'))
            spin_projections = np.transpose(parser.spd[:,:,:,-1, -1], axes=(1,0,2))

            np.save(os.path.join(self.folder, 'spin_projections.npy'), spin_projections)


        spin_projections = spin_projections[:,:,spin]

        if self.hse:
            kpoint_weights = np.array(self.eigenval.kpoints_weights)
            zero_weight = np.where(kpoint_weights == 0)[0]
            spin_projections = spin_projections[:,zero_weight]

        separated_projections = np.ones((spin_projections.shape[0], spin_projections.shape[1], 2), dtype=bool)
        # separated_projections[spin_projections > 0, 0] = spin_projections[spin_projections > 0]
        # separated_projections[spin_projections < 0, 1] = -spin_projections[spin_projections < 0]
        separated_projections[spin_projections > 0, 0] = False
        separated_projections[spin_projections < 0, 1] = False

        # separated_projections = separated_projections / separated_projections.max()
        
        if self.spin == 'up':
            separated_projections = separated_projections[:,:,0]
        elif self.spin == 'down':
            separated_projections = separated_projections[:,:,1]
        else:
            raise BaseException("The soc_axis feature does not work with spin='both'")

        return separated_projections

    def _load_eigenvals(self):
        if self.pre_loaded_bands:
            with open(os.path.join(self.folder, 'eigenvalues.npy'), 'rb') as eigenvals:
                band_data = np.load(eigenvals)

            if self.ispin and not self.lsorbit:
                eigenvalues = band_data[:,:,[0,2]]
                kpoints = band_data[0,:,4:]
                eigenvalues_up = band_data[:,:,[0,1]]
                eigenvalues_down = band_data[:,:,[2,3]]
                if self.spin == 'both':
                    eigenvalues_bg = np.vstack([eigenvalues_up, eigenvalues_down])
                elif self.spin == 'up':
                    eigenvalues_bg = eigenvalues_up
                elif self.spin == 'down':
                    eigenvalues_bg = eigenvalues_down
            else:
                eigenvalues = band_data[:,:,0]
                kpoints = band_data[0,:,2:]
                eigenvalues_bg = band_data[:,:,[0,1]]
        else:
            if len(self.eigenval.eigenvalues.keys()) > 1:
                eigenvalues_up = np.transpose(self.eigenval.eigenvalues[Spin.up], axes=(1,0,2))
                eigenvalues_down = np.transpose(self.eigenval.eigenvalues[Spin.down], axes=(1,0,2))
                eigenvalues_up[:,:,0] = eigenvalues_up[:,:,0] - self.efermi
                eigenvalues_down[:,:,0] = eigenvalues_down[:,:,0] - self.efermi
                eigenvalues = np.concatenate(
                    [eigenvalues_up, eigenvalues_down],
                    axis=2
                )
                if self.spin == 'both':
                    eigenvalues_bg = np.vstack([eigenvalues_up, eigenvalues_down])
                elif self.spin == 'up':
                    eigenvalues_bg = eigenvalues_up
                elif self.spin == 'down':
                    eigenvalues_bg = eigenvalues_down
            else:
                eigenvalues = np.transpose(self.eigenval.eigenvalues[Spin.up], axes=(1,0,2))
                eigenvalues[:,:,0] = eigenvalues[:,:,0] - self.efermi
                eigenvalues_bg = eigenvalues

            kpoints = np.array(self.eigenval.kpoints)

            if self.hse:
                kpoint_weights = np.array(self.eigenval.kpoints_weights)
                zero_weight = np.where(kpoint_weights == 0)[0]
                eigenvalues = eigenvalues[:,zero_weight]
                eigenvalues_bg = eigenvalues_bg[:, zero_weight]
                kpoints = kpoints[zero_weight]

            band_data = np.append(
                eigenvalues,
                np.tile(kpoints, (eigenvalues.shape[0],1,1)),
                axis=2,
            )
            np.save(os.path.join(self.folder, 'eigenvalues.npy'), band_data)

        return eigenvalues_bg

    def _method_0(self, eigenvalues):
        if len(eigenvalues.shape) == 3:
            eigenvalues = eigenvalues[:,:,0]

        occupied = eigenvalues[np.where(eigenvalues < 0)]
        unoccupied = eigenvalues[np.where(eigenvalues > 0)]

        vbm = np.nanmax(occupied)
        cbm = np.nanmin(unoccupied)

        if np.nansum(np.abs(np.diff(np.sign(eigenvalues))) > 0) == 0:
            bg = cbm - vbm
        else:
            bg = 0

        return bg, vbm, cbm

    def _method_1(self, eigenvalues):
        if len(eigenvalues.shape) == 3:
            eigenvalues = eigenvalues[:,:,0]

        band_mean = np.nanmean(eigenvalues, axis=1)

        below_index = np.where(band_mean < 0)[0]
        above_index = np.where(band_mean >= 0)[0]

        vbm = np.nanmax(eigenvalues[below_index])
        cbm = np.nanmin(eigenvalues[above_index])

        if np.nansum(np.abs(np.diff(np.sign(eigenvalues))) > 0) == 0:
            bg = cbm - vbm
        else:
            bg = 0

        return bg, vbm, cbm

    def _get_bandgap(self, method=0):
        bg, vbm, cbm = np.nan, np.nan, np.nan
        eigenvalues = self._load_eigenvals()

        if self.lsorbit:
            if self.spin == "both":
                if method == 0:
                    bg, vbm, cbm = self._method_0(eigenvalues)
                elif method == 1:
                    bg, vbm, cbm = self._method_1(eigenvalues)
            else:
                if self.soc_axis is not None:
                    mask = self._load_soc_spin_projection()
                    eigenvalues[mask] = np.nan
                    if method == 0:
                        bg, vbm, cbm = self._method_0(eigenvalues)
                    elif method == 1:
                        bg, vbm, cbm = self._method_1(eigenvalues)
                else:
                    if method == 0:
                        bg, vbm, cbm = self._method_0(eigenvalues)
                    elif method == 1:
                        bg, vbm, cbm = self._method_1(eigenvalues)
        elif self.ispin and not self.lsorbit:
            if method == 0:
                bg, vbm, cbm = self._method_0(eigenvalues)
            elif method == 1:
                bg, vbm, cbm = self._method_1(eigenvalues)
        elif not self.lsorbit and not self.ispin:
            if method == 0:
                bg, vbm, cbm = self._method_0(eigenvalues)
            elif method == 1:
                bg, vbm, cbm = self._method_1(eigenvalues)

        return bg, vbm, cbm


def get_bandgap_old(
    folder,
    printbg=True,
    return_vbm_cbm=False,
    spin='both',
    method=0,
):
    """
    Determines the band gap from a band structure calculation

    Parameters:
        folder (str): Folder that contains the VASP input and outputs files
        printbg (bool): Determines if the band gap value is printed out or not.
        return_vbm_cbm (bool): Determines if the vbm and cbm are returned.
        spin (str): 'both' returns the bandgap for all spins, 'up' returns the 
            bandgap for only the spin up states and 'down' returns the bandgap
            for only the spin down states.
        method (int): method=0 gets the band gap by finding the values closest to
            the fermi level. method=1 gets the band gap based on the average energy
            of each band.

    Returns:
        if return_vbm_cbm is False: The band gap is returned in eV
        if return_vbm_cbm is True: The band gap, vbm, and cbm are returned in eV in that order
    """
    def _get_bandgap_0(eigenvalues, printbg=printbg):

        occupied = eigenvalues[np.where(eigenvalues[:,:,0] < 0)]
        unoccupied = eigenvalues[np.where(eigenvalues[:,:,0] > 0)]

        vbm = np.max(occupied[:,0])
        cbm = np.min(unoccupied[:,0])

        if np.sum(np.diff(np.sign(eigenvalues[:,:,0])) != 0) == 0:
            bg = cbm - vbm
        else:
            bg = 0

        if printbg:
            print(f'Bandgap = {np.round(bg, 3)} eV')
            if return_vbm_cbm:
                print(f'VBM = {np.round(vbm, 3)} eV')
                print(f'CBM = {np.round(cbm, 3)} eV')

        if return_vbm_cbm:
            return bg, vbm, cbm
        else:
            return bg

    def _get_bandgap_1(eigenvalues, printbg=printbg):
        band_mean = eigenvalues[:,:,0].mean(axis=1)

        below_index = np.where(band_mean < 0)[0]
        above_index = np.where(band_mean >= 0)[0]

        vbm = np.max(eigenvalues[below_index, :, 0])
        cbm = np.min(eigenvalues[above_index, :, 0])

        if np.sum(np.diff(np.sign(eigenvalues[:,:,0])) != 0) == 0:
            bg = cbm - vbm
        else:
            bg = 0

        if printbg:
            print(f'Bandgap = {np.round(bg, 3)} eV')
            if return_vbm_cbm:
                print(f'VBM = {np.round(vbm, 3)} eV')
                print(f'CBM = {np.round(cbm, 3)} eV')

        if return_vbm_cbm:
            return bg, vbm, cbm
        else:
            return bg

    pre_loaded_bands = os.path.isfile(os.path.join(folder, 'eigenvalues.npy'))
    eigenval = Eigenval(os.path.join(folder, 'EIGENVAL'))
    efermi = float(os.popen(f'grep E-fermi {os.path.join(folder, "OUTCAR")}').read().split()[2])
    incar = Incar.from_file(
        os.path.join(folder, 'INCAR')
    )
    if 'LSORBIT' in incar:
        if incar['LSORBIT']:
            lsorbit = True
        else:
            lsorbit = False
    else:
        lsorbit = False

    if 'ISPIN' in incar:
        if incar['ISPIN'] == 2:
            ispin = True
        else:
            ispin = False
    else:
        ispin = False

    if 'LHFCALC' in incar:
        if incar['LHFCALC']:
            hse = True
        else:
            hse = False
    else:
        hse = False

    if pre_loaded_bands:
        with open(os.path.join(folder, 'eigenvalues.npy'), 'rb') as eigenvals:
            band_data = np.load(eigenvals)

        if ispin and not lsorbit:
            eigenvalues = band_data[:,:,[0,2]]
            kpoints = band_data[0,:,4:]
            eigenvalues_up = band_data[:,:,[0,1]]
            eigenvalues_down = band_data[:,:,[2,3]]
            if spin == 'both':
                eigenvalues_bg = np.vstack([eigenvalues_up, eigenvalues_down])
            elif spin == 'up':
                eigenvalues_bg = eigenvalues_up
            elif spin == 'down':
                eigenvalues_bg = eigenvalues_down
        else:
            eigenvalues = band_data[:,:,0]
            kpoints = band_data[0,:,2:]
            eigenvalues_bg = band_data[:,:,[0,1]]
        
        if method == 0:
            band_gap = _get_bandgap_0(eigenvalues=eigenvalues_bg)
        elif method == 1:
            band_gap = _get_bandgap_1(eigenvalues=eigenvalues_bg)

    else:
        if len(eigenval.eigenvalues.keys()) > 1:
            eigenvalues_up = np.transpose(eigenval.eigenvalues[Spin.up], axes=(1,0,2))
            eigenvalues_down = np.transpose(eigenval.eigenvalues[Spin.down], axes=(1,0,2))
            eigenvalues_up[:,:,0] = eigenvalues_up[:,:,0] - efermi
            eigenvalues_down[:,:,0] = eigenvalues_down[:,:,0] - efermi
            eigenvalues = np.concatenate(
                [eigenvalues_up, eigenvalues_down],
                axis=2
            )
            if spin == 'both':
                eigenvalues_bg = np.vstack([eigenvalues_up, eigenvalues_down])
            elif spin == 'up':
                eigenvalues_bg = eigenvalues_up
            elif spin == 'down':
                eigenvalues_bg = eigenvalues_down
        else:
            eigenvalues = np.transpose(eigenval.eigenvalues[Spin.up], axes=(1,0,2))
            eigenvalues[:,:,0] = eigenvalues[:,:,0] - efermi
            eigenvalues_bg = eigenvalues

        kpoints = np.array(eigenval.kpoints)

        if hse:
            kpoint_weights = np.array(eigenval.kpoints_weights)
            zero_weight = np.where(kpoint_weights == 0)[0]
            eigenvalues = eigenvalues[:,zero_weight]
            eigenvalues_bg = eigenvalues_bg[:, zero_weight]
            kpoints = kpoints[zero_weight]

        if method == 0:
            band_gap = _get_bandgap_0(eigenvalues=eigenvalues_bg)
        elif method == 1:
            band_gap = _get_bandgap_1(eigenvalues=eigenvalues_bg)

        band_data = np.append(
            eigenvalues,
            np.tile(kpoints, (eigenvalues.shape[0],1,1)),
            axis=2,
        )

        np.save(os.path.join(folder, 'eigenvalues.npy'), band_data)

    return band_gap


def passivator(
    struc,
    passivated_struc=None,
    top=True,
    bot=True,
    symmetrize=True,
    tol=0.0001,
    write_file=False,
    output='POSCAR_pas',
):
    """
    This function passivates the slabs with pseudohydrogen. The positions of the pseudo-
    hydrogen are determined by locating the bonds between the second to last and last
    layers and places hydrogen along each bond at a distance proportional to the sum of
    the covalent radii of the given species and the passivating hydrogen. If an already passivated
    structure is input, the location of the passivating layer will be used on the new structure.

    Parameters:
        struc (str or pymatgen.core.structure.Structure): Unpassivated Structure or file path to
            unpassivated structure.
        passivated_struc (str or pymatgen.core.Structure): Structure or file path to a structure whos
            passivation layer has already been relaxed
        top (bool): Determines if the top of the slab is passivated
        bot (bool): Determines if the bottom of the slab is passivated
        symmetrize (bool): Determines if the slab is symmetrized or not
        tol (float): Numerical tolerence for determining the atoms that get grouped
            into an atomic layer.

    Returns:
        struc_pas (pymatgen.core.Structure): Passivated pymatgen structure
    """

    if type(struc) == str:
        struc = Structure.from_file(struc)
    if type(passivated_struc) == str:
        passivated_struc = Structure.from_file(passivated_struc)

    struc, shift = _center_slab(struc)

    sorted_slab, z_positions = _sort_by_z(struc)

    if passivated_struc is None:
        sorted_slab2, z_positions2 = _sort_by_z(struc)
    else:
        sorted_slab2, z_positions2 = _sort_by_z(passivated_struc)

    for i in range(len(sorted_slab)):
        sorted_slab.sites[i].properties = {'to_delete': False}

    for i in range(len(sorted_slab2)):
        sorted_slab2.sites[i].properties = {'to_delete': False}

    top_index = _get_top_index(z_positions, tol=tol)

    second_top_index = _get_top_index(
        z_positions, to_delete=top_index, tol=tol
    )

    bot_index = _get_bot_index(z_positions, tol=tol)

    if symmetrize:
        z_positions = np.delete(z_positions, bot_index)

        for i in bot_index:
            del sorted_slab[-1]

        bot_index = _get_bot_index(z_positions, tol=tol)

    second_bot_index = _get_bot_index(
        z_positions, to_delete=bot_index, tol=tol
    )

    if passivated_struc is None:
        top_index2 = _get_top_index(z_positions2, tol=tol)

        second_top_index2 = _get_top_index(
            z_positions2, to_delete=top_index2, tol=tol
        )

        bot_index2 = _get_bot_index(z_positions2, tol=tol)

        if symmetrize:
            z_positions2 = np.delete(z_positions2, bot_index2)

            for i in bot_index:
                del sorted_slab2[-1]

            bot_index2 = _get_bot_index(z_positions2, tol=tol)

        second_bot_index2 = _get_bot_index(
            z_positions2, to_delete=bot_index2, tol=tol
        )

    else:
        H_index = np.where(
            np.array([str(i) for i in sorted_slab2.species]) == 'H'
        )[0]

        second_top_index2 = _get_top_index(
            z_positions2, to_delete=H_index, tol=tol
        )

        second_bot_index2 = _get_bot_index(
            z_positions2, to_delete=H_index, tol=tol
        )

        top_index2 = range(np.min(second_top_index2))

        bot_index2 = range(np.max(second_bot_index2) +
                           1, len(z_positions2))

    if bot:
        for i in bot_index:
            sorted_slab.sites[i].properties['to_delete'] = True

        for i in bot_index2:
            sorted_slab2.sites[i].properties['to_delete'] = True

    if top:
        for i in top_index:
            sorted_slab.sites[i].properties['to_delete'] = True

        for i in top_index2:
            sorted_slab2.sites[i].properties['to_delete'] = True

    elements = [str(element) for element in np.unique(sorted_slab.species)]

    max_covalent_radius = np.max(
        [CovalentRadius.radius[element] for element in elements]
    )

    if passivated_struc is None:
        new_radius = True
    else:
        new_radius = False

    if bot:
        for (i, j) in zip(second_bot_index, second_bot_index2):
            neighbor_sph_coords = _get_neighbors(
                struc=sorted_slab2,
                index=j,
                covalent_radius=max_covalent_radius
            )

            _append_H(
                struc=sorted_slab,
                index=i,
                neighbor_sph_coords=neighbor_sph_coords,
                side='bot',
                new_radius=new_radius,
            )

        sorted_slab, _ = _sort_by_z(sorted_slab)
        sorted_slab2, _ = _sort_by_z(sorted_slab2)

    if top:
        for (i, j) in zip(second_top_index, second_top_index2):
            neighbor_sph_coords = _get_neighbors(
                struc=sorted_slab2,
                index=j,
                covalent_radius=max_covalent_radius
            )

            _append_H(
                struc=sorted_slab,
                index=i,
                neighbor_sph_coords=neighbor_sph_coords,
                side='top',
                new_radius=new_radius,
            )

    sites_to_delete = [
        i for i in range(len(sorted_slab)) if sorted_slab.sites[i].properties['to_delete']
    ]
    sorted_slab.remove_sites(sites_to_delete)

    sorted_slab = sorted_slab.get_sorted_structure()
    sorted_slab.translate_sites(range(len(sorted_slab)), [0, 0, -shift])

    if write_file:
        Poscar(sorted_slab).write_file(output)

    return sorted_slab

def get_periodic_vacuum(
        slab,
        bulk,
        miller_index,
        vacuum=40,
        write_file=False,
        output='POSCAR_good_vacuum',
        periodic_vacuum=True,
):
    """
    Creates a slab with a vacuum that is an integer multiple of unit cell length in the direction of the
    miller index. This is necessary for proper band unfolding.

    Parameters:
        slab (str or pymatgen.core.structure.Structure): File path to a slab structure or a pymatgen Structure
        bulk (str or pymatgen.core.structure.Structure): File path to a bulk structure or a pymatgen Structure
        miller_index (list): 3 element list for the miller index of the surface
        vacuum (float): Size of the vacuum. This will be rounded to the closest integer multiple of the unit
            cell length in the direction of the given miller index
        write_file (bool): Determines if a POSCAR is written
        output (str): File name of output POSCAR

    Returns:
        slab structure that has a vacuum with the correct size for band unfolding
    """
    if type(slab) == str:
        slab = Structure.from_file(slab)
    if type(bulk) == str:
        bulk = Structure.from_file(bulk)

    index = np.array(miller_index).reshape(1,-1)
    metric_tensor = bulk.lattice.metric_tensor
    unit_cell_len = np.sqrt(np.squeeze(np.matmul(index,np.matmul(metric_tensor, index.T))))
    c_len = np.linalg.norm(slab.lattice.matrix[-1])

    if np.isin(np.array(slab.species, dtype=str), 'H').any():
        inds = np.isin(np.array(slab.species, dtype=str), 'H', invert=True)
        min_z = np.min(slab.frac_coords[inds,-1])
        slab.translate_sites(
            range(len(slab)),
            [0,0,((unit_cell_len / c_len) - min_z)],
        )
    else:
        min_z = np.min(slab.frac_coords[:,-1])
        slab.translate_sites(
            range(len(slab)),
            [0,0, -min_z],
        )

    max_z2 = np.max(slab.frac_coords[:,-1])
    min_z2 = np.min(slab.frac_coords[:,-1])
    slab_height_in_unit_cells = np.round(((max_z2 - min_z2) * c_len) / unit_cell_len, 0)
    vacuum_height_in_unit_cells = np.round(vacuum / unit_cell_len, 0)
    new_lattice = copy.deepcopy(slab.lattice.matrix)

    if periodic_vacuum:
        new_lattice[-1] = (new_lattice[-1] / np.linalg.norm(new_lattice[-1])) * (slab_height_in_unit_cells + vacuum_height_in_unit_cells) * unit_cell_len
    else:
        new_lattice[-1] = (new_lattice[-1] / np.linalg.norm(new_lattice[-1])) * (((max_z2 - min_z2) * c_len) + vacuum)

    new_c_len = np.linalg.norm(new_lattice[-1])
    new_frac_coords = copy.deepcopy(slab.frac_coords)
    new_frac_coords[:,-1] = new_frac_coords[:,-1] * (c_len / new_c_len)

    new_structure = Structure(
        lattice=Lattice(new_lattice),
        species=slab.species,
        coords=new_frac_coords,
        to_unit_cell=True,
    )

    if write_file:
        Poscar(new_structure).write_file(output)
    else:
        return new_structure



def make_supercell(slab, scaling_matrix):
    """
    Generates a supercell given a scaling matrix

    Parameters:
        slab (str or pymatgen.core.structure.Structure): File path to a slab structure or a pymatgen Structure
        scaling_matrix (list): 3 element scaling matrix

    Returns:
        Super cell slab
    """
    if type(slab) == str:
        slab = Structure.from_file(slab)

    supercell_slab = copy.deepcopy(slab)
    supercell_slab.make_supercell(scaling_matrix=scaling_matrix)

    return supercell_slab

def generate_slab(
        bulk,
        miller_index,
        layers,
        vacuum,
        scaling_matrix=None,
        write_file=True,
        output=None,
        passivate=False,
        passivated_file=None,
        passivate_top=True,
        passivate_bot=True,
        symmetrize=False,
        tol=0.0001,
        periodic_vacuum=True,
):
    """
    This function generates a slab structure.

    Parameters:
        bulk_structure (str): File path to the POSCAR of the conventional 
            unit cell.
        miller_index (list): Three element list to define the miller indices
        layers (int): Number of unit layers in the slab structure
        vacuum (float): Size of the vacuum in Angstoms.
        scaling_matrix (list): 3 element scaling matrix
        write_file (bool): Determines if a POSCAR is written
        output (str): File name of output POSCAR
        passivated_file (pymatgen.core.Structure): path to structure whos passivation layer has already been relaxed
        top (bool): Determines if the top of the slab is passivated
        bot (bool): Determines if the bottom of the slab is passivated
        symmetrize (bool): Determines if the slab is symmetrized (top and bottom have the same termination) or not
        tol (float): Numerical tolerence for determining the atoms that get grouped
            into an atomic layer.

    Returns: 
        Slab structure 
    """

    bulk_pmg = Structure.from_file(bulk)
    bulk_sg = SpacegroupAnalyzer(bulk_pmg)
    bulk_structure_conv = bulk_sg.get_conventional_standard_structure()
    bulk_structure_prim = bulk_sg.get_primitive_standard_structure()
    bulk_structure_ase = AseAtomsAdaptor().get_atoms(bulk_structure_conv)

    ase_slab = surface(
        bulk_structure_ase,
        miller_index,
        layers,
        vacuum=40,
        periodic=False
    )
    sorted_slab = sort(ase_slab, tags=ase_slab.positions[:,-1])
    pmg_structure = AseAtomsAdaptor().get_structure(sorted_slab)
    pmg_sorted_slab = pmg_structure.get_sorted_structure()
    slab_space = SpacegroupAnalyzer(pmg_sorted_slab)
    slab_primitive = slab_space.get_primitive_standard_structure()
    slab_primitive = get_periodic_vacuum(
        slab_primitive,
        bulk_structure_prim,
        miller_index,
        vacuum=vacuum,
        periodic_vacuum=periodic_vacuum,
    )

    if passivate:
        if passivated_file is not None:
            passivated_file = Structure.from_file(passivated_file)

        slab_primitive = passivator(
            slab_primitive,
            passivated_struc=passivated_file,
            top=passivate_top,
            bot=passivate_bot,
            symmetrize=symmetrize,
            tol=tol,
        )

    if scaling_matrix is not None:
        slab_primitive.make_supercell(scaling_matrix=scaling_matrix)

    if write_file:
        if output is None:
            Poscar(slab_primitive).write_file(f'POSCAR_{layers}', direct=True)
        else:
            Poscar(slab_primitive).write_file(output, direct=True)

    return slab_primitive


def compare_dos_to_bulk(
    bulk_folder,
    slab_folder,
    atoms,
    erange=[-6,6],
    sigma=0.05,
    spin_polarized=False,
    plot=True,
    save_plot=True,
    output='bulk_comparison.png',
    bulk_color='black',
    slab_color='green',
    fill_bulk=True,
    fill_slab=True,
    figsize=(4,3),
):
    from vaspvis.dos import Dos
    from vaspvis.standard import _figure_setup_dos
    fig, ax = plt.subplots(figsize=figsize, dpi=300)
    _figure_setup_dos(ax=ax, fontsize=12, energyaxis='x')

    if spin_polarized:
        dos_bulk_up = Dos(
            folder=bulk_folder,
            spin='up',
        )
        dos_bulk_down = Dos(
            folder=bulk_folder,
            spin='down',
        )

        dos_bulk_up.plot_plain(
            ax=ax,
            energyaxis='x',
            color=bulk_color,
            erange=erange,
            fill=fill_bulk,
        )
        dos_bulk_down.plot_plain(
            ax=ax,
            energyaxis='x',
            color=bulk_color,
            erange=erange,
            fill=fill_bulk,
        )

        dos_slab_up = Dos(
            folder=slab_folder,
            spin='up',
        )

        dos_slab_down = Dos(
            folder=slab_folder,
            spin='down',
        )

        dos_slab_up.plot_atoms(
            ax=ax,
            atoms=atoms,
            legend=False,
            energyaxis='x',
            color_list=[slab_color for _ in range(len(atoms))],
            total=False,
            fill=fill_slab,
            sum_atoms=True,
            erange=erange,
        )

        dos_slab_down.plot_atoms(
            ax=ax,
            atoms=atoms,
            legend=False,
            energyaxis='x',
            color_list=[slab_color for _ in range(len(atoms))],
            total=False,
            fill=fill_slab,
            sum_atoms=True,
            erange=erange,
        )

        x_data = [ax.lines[i].get_xdata() for i in range(4)]
        y_data = [ax.lines[i].get_ydata() for i in range(4)]

        area_diff_up = fastdtw(
            np.c_[x_data[0], y_data[0]],
            np.c_[x_data[2], y_data[2]],
        )[0]
        area_diff_down = fastdtw(
            np.c_[x_data[1], y_data[1]],
            np.c_[x_data[3], y_data[3]],
        )[0]
        area_diff = np.abs(area_diff_up) + np.abs(area_diff_down)
    else:
        ax.set_xlim(erange[0], erange[1])

        dos_bulk = Dos(
            folder=bulk_folder,
        )

        dos_bulk.plot_plain(
            ax=ax,
            energyaxis='x',
            color=bulk_color,
            erange=erange,
            fill=fill_bulk,
        )

        dos = Dos(
            folder=slab_folder,
        )

        dos.plot_atoms(
            ax=ax,
            atoms=atoms,
            legend=False,
            energyaxis='x',
            color_list=[slab_color for _ in range(len(atoms))],
            total=False,
            fill=fill_slab,
            sum_atoms=True,
            erange=erange,
        )
        x_data = [ax.lines[i].get_xdata() for i in range(2)]
        y_data = [ax.lines[i].get_ydata() for i in range(2)]

        area_diff = fastdtw(
            np.c_[x_data[0], y_data[0]],
            np.c_[x_data[1], y_data[1]],
        )[0]

        area_diff = np.abs(area_diff)

    if plot:
        fig.tight_layout()
        if save_plot:
            fig.savefig(output)

            return area_diff
        else:
            return fig, ax, area_diff
    else:
        return area_diff




if __name__ == "__main__":
    #  area_diff = compare_dos_to_bulk(
        #  bulk_folder='../../vaspvis_data/dos_InAs/',
        #  slab_folder='../../vaspvis_data/slabdos',
        #  atoms=[17,50],
    #  )
    #  slab = passivator(
        #  struc=Poscar.from_file('./POSCAR_0').structure,
        #  write_file=True,
    #  )
    #  Poscar(slab).write_file('POSCAR_pas')
    #  M = convert_slab(
        #  bulk_path='../../../../projects/unfold_test/POSCAR_bulk',
        #  slab_path=slab,
        #  index=[1,1,1],
    #  )
    #  high_symmetry_points = [
        #  [0.1, 0.1, 0.0],
        #  [0.0, 0.0, 0.0],
        #  [0.1, 0.1, 0.0],
    #  ]
    #  generate_kpoints(
        #  M=M,
        #  high_symmetry_points=high_symmetry_points,
        #  n=40,
        #  output='KPOINTS_AGA',
    #  )
    #  get_periodic_vacuum(
        #  slab='./POSCAR_30',
        #  bulk='../../../../projects/unfold_test/POSCAR_bulk',
        #  miller_index=[1,1,1],
        #  vacuum=40,
        #  write_file=True
    #  )
    # slab = passivator(
    #         struc=Poscar.from_file('./POSCAR_2').structure,
    #         # struc='POSCAR_pas',
    #         write_file=True,
    #         output='POSCAR_transfer',
    #         passivated_struc='./CONTCAR',
    #         symmetrize=False,
    #         tol=0.005,
    #         )
    gap_both = BandGap(
            folder='../../vaspvis_data/band_InAs',
            spin='both',
            soc_axis=None
            )
    print('Both')
    print('Gap =', np.round(gap_both.bg, 3))
    print('VBM =', np.round(gap_both.vbm, 3))
    print('CBM =', np.round(gap_both.cbm, 3))
    # print('')
    # gap_up = BandGap(folder='../../vaspvis_data/Ti2MnIn_band', spin='up', soc_axis='z')
    # print('Up')
    # print('Gap =', np.round(gap_up.bg, 3))
    # print('VBM =', np.round(gap_up.vbm, 3))
    # print('CBM =', np.round(gap_up.cbm, 3))
    # print('')
    # gap_down = BandGap(folder='../../vaspvis_data/Ti2MnIn_band', spin='down', soc_axis='z')
    # print('Down')
    # print('Gap =', np.round(gap_down.bg, 3))
    # print('VBM =', np.round(gap_down.vbm, 3))
    # print('CBM =', np.round(gap_down.cbm, 3))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

############################################################
import numpy as np
import multiprocessing
from .vaspwfc import vaspwfc

############################################################


def find_K_from_k(k, M):
    """
    Get the K vector of the supercell onto which the k vector of the primitive
    cell folds. The unfoliding vector G, which satisfy the following equation,
    is also returned.

        k = K + G

    where G is a reciprocal space vector of supercell
    """

    M = np.array(M)
    Kc = np.dot(k, M.T)
    G = np.array(np.round(Kc), dtype=int)
    KG = Kc - np.round(Kc)

    return KG, G


def LorentzSmearing(x, x0, sigma=0.02):
    """
    Simulate the Delta function by a Lorentzian shape function

        \Delta(x) = \lim_{\sigma\to 0}  Lorentzian
    """

    return 1.0 / np.pi * sigma**2 / ((x - x0) ** 2 + sigma**2)


def GaussianSmearing(x, x0, sigma=0.02):
    """
    Simulate the Delta function by a Lorentzian shape function

        \Delta(x) = \lim_{\sigma\to 0} Gaussian
    """

    return (
        1.0
        / (np.sqrt(2 * np.pi) * sigma)
        * np.exp(-((x - x0) ** 2) / (2 * sigma**2))
    )


def removeDuplicateKpoints(kpoints):
    """
    remove duplicate kpoints in the list.
    """
    kpoints = np.array(sorted(kpoints, key=lambda x: (x[0], x[1], x[2])))
    kdiff = np.diff(kpoints, axis=0)
    match = np.abs(np.linalg.norm(kdiff, axis=1)) > 1e-5

    return kpoints[np.r_[True, match]]


def save2VaspKPOINTS(kpoints, output="KPOINTS"):
    """
    save to VASP KPOINTS file
    """
    kpoints = np.array(kpoints)
    nkpts = kpoints.shape[0]

    with open(output, "w") as vaspkpt:
        vaspkpt.write("kpoints generated by PYVASPWFC\n")
        vaspkpt.write("%d\n" % nkpts)
        vaspkpt.write("Rec\n")
        for ik in range(nkpts):
            line = "  %12.8f %12.8f %12.8f 1.0\n" % (
                kpoints[ik, 0],
                kpoints[ik, 1],
                kpoints[ik, 2],
            )
            vaspkpt.write(line)


def make_kpath(kbound, nseg=40):
    """
    make a list of kpoints defining the path between the given kpoints.
    """
    kbound = np.array(kbound, dtype=float)
    kdist = np.diff(kbound, axis=0)

    # kpath = []
    # for ii in range(len(kdist)):
    #     for nk in range(nseg):
    #         kpt = kbound[ii] + kdist[ii] / nseg * nk
    #         kpath.append(kpt)

    kpath = [
        kbound[ii] + kdist[ii] / nseg * nk
        for ii in range(len(kdist))
        for nk in range(nseg)
    ]
    kpath.append(kbound[-1])
    return kpath


def EBS_scatter(
    kpts,
    cell,
    spectral_weight,
    eref=0.0,
    nseg=None,
    save="ebs_s.png",
    kpath_label=[],
    factor=20,
    figsize=(3.0, 4.0),
    ylim=(-3, 3),
    show=True,
    color="b",
):
    """
    plot the effective band structure with scatter, the size of the scatter
    indicates the spectral weight.
    The plotting function utilizes Matplotlib package.

    inputs:
        kpts: the kpoints vectors in fractional coordinates.
        cell: the primitive cell basis
        spectral_weight: self-explanatory
    """

    import matplotlib as mpl

    mpl.use("agg")
    import matplotlib.pyplot as plt

    mpl.rcParams["axes.unicode_minus"] = False

    nspin = spectral_weight.shape[0]
    kpt_c = np.dot(kpts, np.linalg.inv(cell).T)
    kdist = np.r_[0, np.cumsum(np.linalg.norm(np.diff(kpt_c, axis=0), axis=1))]
    nk = kdist.size
    nb = spectral_weight.shape[2]
    x0 = np.ones(nb)

    fig = plt.figure()
    fig.set_size_inches(figsize)
    if nspin == 1:
        axes = [plt.subplot(111)]
        fig.set_size_inches(figsize)
    else:
        axes = [plt.subplot(121), plt.subplot(122)]
        fig.set_size_inches((figsize[0] * 2, figsize[1]))

    for ispin in range(nspin):
        ax = axes[ispin]
        for ik in range(nk):
            ax.scatter(
                kdist[ik] * x0,
                spectral_weight[ispin, ik, :, 0] - eref,
                s=spectral_weight[ispin, ik, :, 1] * factor,
                lw=0.0,
                #  color=color
            )

        ax.set_xlim(0, kdist.max())
        ax.set_ylim(*ylim)
        ax.set_ylabel("Energy [eV]", labelpad=5)

        if nseg:
            for kb in kdist[::nseg]:
                ax.axvline(x=kb, lw=0.5, color="k", ls=":", alpha=0.8)

            if kpath_label:
                ax.set_xticks(np.r_[kdist[::nseg], kdist[-1]])
                kname = [x.upper() for x in kpath_label]
                for ii in range(len(kname)):
                    if kname[ii] == "G":
                        kname[ii] = r"$\mathrm{\mathsf{\Gamma}}$"
                    else:
                        kname[ii] = r"$\mathrm{\mathsf{%s}}$" % kname[ii]
                ax.set_xticklabels(kname)

    plt.tight_layout(pad=0.2)
    plt.savefig(save, dpi=360)
    if show:
        plt.show()


def EBS_cmaps(
    kpts,
    cell,
    E0,
    spectral_function,
    eref=0.0,
    nseg=None,
    kpath_label=[],
    save="ebs_c.png",
    figsize=(3.0, 4.0),
    ylim=(-3, 3),
    show=True,
    cmap="jet",
):
    """
    plot the effective band structure with colormaps.  The plotting function
    utilizes Matplotlib package.

    inputs:
        kpts: the kpoints vectors in fractional coordinates.
        cell: the primitive cell basis
        spectral_weight: self-explanatory
    """

    import matplotlib as mpl

    mpl.use("agg")
    import matplotlib.pyplot as plt

    mpl.rcParams["axes.unicode_minus"] = False

    nspin = spectral_function.shape[0]
    kpt_c = np.dot(kpts, np.linalg.inv(cell).T)
    kdist = np.r_[0, np.cumsum(np.linalg.norm(np.diff(kpt_c, axis=0), axis=1))]
    nk = kdist.size
    xmin, xmax = kdist.min(), kdist.max()
    # ymin, ymax = E0.min() - eref, E0.max() - eref

    fig = plt.figure()
    if nspin == 1:
        axes = [plt.subplot(111)]
        fig.set_size_inches(figsize)
    else:
        axes = [plt.subplot(121), plt.subplot(122)]
        fig.set_size_inches((figsize[0] * 2, figsize[1]))

    # ax.imshow(spectral_function, extent=(xmin, xmax, ymin, ymax),
    #           origin='lower', aspect='auto', cmap=cmap)
    X, Y = np.meshgrid(kdist, E0 - eref)
    for ispin in range(nspin):
        ax = axes[ispin]
        ax.contourf(X, Y, spectral_function[ispin], cmap=cmap)

        ax.set_xlim(xmin, xmax)
        ax.set_ylim(*ylim)
        ax.set_ylabel("Energy [eV]", labelpad=5)

        if nseg:
            for kb in kdist[::nseg]:
                ax.axvline(x=kb, lw=0.5, color="k", ls=":", alpha=0.8)

            if kpath_label:
                ax.set_xticks(np.r_[kdist[::nseg], kdist[-1]])
                kname = [x.upper() for x in kpath_label]
                for ii in range(len(kname)):
                    if kname[ii] == "G":
                        kname[ii] = r"$\mathrm{\mathsf{\Gamma}}$"
                    else:
                        kname[ii] = r"$\mathrm{\mathsf{%s}}$" % kname[ii]
                ax.set_xticklabels(kname)

    plt.tight_layout(pad=0.2)
    plt.savefig(save, dpi=360)
    if show:
        plt.show()


############################################################


class unfold:
    """
    Unfold the band structure from Supercell calculation into a primitive cell and
    obtain the effective band structure (EBS).

    REF:
    "Extracting E versus k effective band structure from supercell
     calculations on alloys and impurities"
    Phys. Rev. B 85, 085201 (2012)
    """

    def __init__(
        self,
        M=None,
        wavecar="WAVECAR",
        gamma=False,
        lsorbit=False,
        gamma_half="x",
    ):
        """
        Initialization.

        M is the transformation matrix between supercell and primitive cell:

            M = np.dot(A, np.linalg.inv(a))

        In real space, the basis vectors of Supercell (A) and those of the
        primitive cell (a) satisfy:

            A = np.dot(M, a);      a = np.dot(np.linalg.inv(M), A)

        Whereas in reciprocal space

            b = np.dot(M.T, B);    B = np.dot(np.linalg.inv(M).T, b)

        wavecar is the location of VASP WAVECAR file that contains the
        wavefunction information of a supercell calculation.
        """

        # Whether the WAVECAR is a gamma-only version
        self._lgam = gamma
        self._lsoc = lsorbit

        self.M = np.array(M, dtype=float)
        assert self.M.shape == (
            3,
            3,
        ), "Shape of the tranformation matrix must be (3,3)"

        self.wfc = vaspwfc(
            wavecar,
            lsorbit=self._lsoc,
            lgamma=self._lgam,
            gamma_half=gamma_half,
        )
        # all the K-point vectors
        self.kvecs = self.wfc._kvecs
        # all the KS energies
        self.bands = self.wfc._bands
        #  print(self.bands[0][15])

        # G-vectors within the cutoff sphere, let's just do it once for all.
        # self.allGvecs = np.array([self.wfc.gvectors(ikpt=kpt+1)
        #                           for kpt in range(self.wfc._nkpts)], dtype=int)

        # spectral weight for all the kpoints
        self.SW = None

    def get_ovlap_G(self, ikpt=1, epsilon=1e-5):
        """
        Get subset of the reciprocal space vectors of the supercell,
        specifically the ones that match the reciprocal space vectors of the
        primitive cell.
        """

        assert 1 <= ikpt <= self.wfc._nkpts, "Invalid K-point index!"

        # Reciprocal space vectors of the supercell in fractional unit
        Gvecs = self.wfc.gvectors(ikpt=ikpt)
        # Gvecs = self.allGvecs[ikpt - 1]

        if self._lgam:
            nplw = Gvecs.shape[0]
            tmp = np.zeros((nplw * 2 - 1, 3), dtype=int)
            # the gvectors of Gamma version only contains half the number of a
            # normal version.
            tmp[:nplw, ...] = Gvecs
            tmp[nplw:, ...] = -Gvecs[1:, ...]  # G' = -G

            Gvecs = tmp

        # Shape of Gvecs: (nplws, 3)
        # iGvecs = np.arange(Gvecs.shape[0], dtype=int)

        # Reciprocal space vectors of the primitive cell
        gvecs = np.dot(Gvecs, np.linalg.inv(self.M).T)
        # Deviation from the perfect sites
        gd = gvecs - np.round(gvecs)
        # match = np.linalg.norm(gd, axis=1) < epsilon
        match = np.alltrue(np.abs(gd) < epsilon, axis=1)

        # return Gvecs[match], iGvecs[match]
        return Gvecs[match], Gvecs

    def find_K_index(self, K0):
        """
        Find the index of K0.
        """

        for ii in range(self.wfc._nkpts):
            if np.alltrue(np.abs(self.wfc._kvecs[ii] - K0) < 1e-5):
                return ii + 1
        # the for-else
        else:
            raise ValueError(
                "Cannot find the corresponding K-points in WAVECAR!"
            )

    def k2K_map(self, kpath):
        """
        Find the map from primitive-cell k-points to supercell k-points.
        """

        return [
            self.find_K_index(find_K_from_k(k, self.M)[0]) - 1 for k in kpath
        ]

    def spectral_weight_k(self, k0, whichspin=1):
        """
        Spectral weight for a given k:

            P_{Km}(k) = \sum_n |<Km | kn>|^2

        which is equivalent to

            P_{Km}(k) = \sum_{G} |C_{Km}(G + k - K)|^2

        where {G} is a subset of the reciprocal space vectors of the supercell.
        """

        #  print('Processing k-point %8.4f %8.4f %8.4f' % (k0[0], k0[1], k0[2]))

        # find the K0 onto which k0 folds
        # k0 = G0 + K0
        K0, G0 = find_K_from_k(k0, self.M)
        # find index of K0
        ikpt = self.find_K_index(K0)

        # get the overlap G-vectors
        Gvalid, Gall = self.get_ovlap_G(ikpt=ikpt)
        # Gnew = Gvalid + k0 - K0
        Goffset = Gvalid + G0[np.newaxis, :]

        # Index of the Gvalid in 3D grid
        GallIndex = Gall % self.wfc._ngrid[np.newaxis, :]
        GoffsetIndex = Goffset % self.wfc._ngrid[np.newaxis, :]

        # 3d grid for planewave coefficients
        wfc_k_3D = np.zeros(self.wfc._ngrid, dtype=np.complex128)

        # if self._lsoc:
        #     # the weights and corresponding energies
        #     P_Km = np.zeros((2, self.wfc._nbands), dtype=float)
        #     E_Km = np.zeros((2, self.wfc._nbands), dtype=float)
        # else:
        #     # the weights and corresponding energies
        #     P_Km = np.zeros(self.wfc._nbands, dtype=float)
        #     E_Km = np.zeros(self.wfc._nbands, dtype=float)

        # the weights and corresponding energies
        P_Km = np.zeros(self.wfc._nbands, dtype=float)
        E_Km = np.zeros(self.wfc._nbands, dtype=float)

        for nb in range(self.wfc._nbands):
            # initialize the array to zero, which is unnecessary since the
            # GallIndex is the same for the same K-point
            # wfc_k_3D[:,:,:] = 0.0

            if self._lsoc:
                # pad the coefficients to 3D grid
                band_coeff = self.wfc.readBandCoeff(
                    ispin=whichspin, ikpt=ikpt, iband=nb + 1, norm=True
                )
                nplw = band_coeff.shape[0] // 2
                band_spinor_coeff = [band_coeff[:nplw], band_coeff[nplw:]]

                # energy
                E_Km[nb] = self.bands[whichspin - 1, ikpt - 1, nb]
                for Ispinor in range(2):
                    # band = band_spinor_coeff[Ispinor]
                    # band /= np.linalg.norm(band)
                    wfc_k_3D[
                        GallIndex[:, 0], GallIndex[:, 1], GallIndex[:, 2]
                    ] = band_spinor_coeff[Ispinor]

                    # spectral weight
                    P_Km[nb] += (
                        np.linalg.norm(
                            wfc_k_3D[
                                GoffsetIndex[:, 0],
                                GoffsetIndex[:, 1],
                                GoffsetIndex[:, 2],
                            ]
                        )
                        ** 2
                    )
            else:
                # pad the coefficients to 3D grid
                band_coeff = self.wfc.readBandCoeff(
                    ispin=whichspin, ikpt=ikpt, iband=nb + 1, norm=True
                )
                if self._lgam:
                    nplw = band_coeff.size
                    tmp = np.zeros((nplw * 2 - 1), dtype=band_coeff.dtype)
                    # for Gamma version, the coefficients corresponding to G \ne 0
                    # is multiplied by a factor of sqrt(2)
                    band_coeff[1:] /= np.sqrt(2.0)
                    tmp[:nplw] = band_coeff
                    tmp[nplw:] = band_coeff[1:].conj()
                    band_coeff = tmp

                wfc_k_3D[
                    GallIndex[:, 0], GallIndex[:, 1], GallIndex[:, 2]
                ] = band_coeff
                # energy
                E_Km[nb] = self.bands[whichspin - 1, ikpt - 1, nb]
                # spectral weight
                P_Km[nb] = (
                    np.linalg.norm(
                        wfc_k_3D[
                            GoffsetIndex[:, 0],
                            GoffsetIndex[:, 1],
                            GoffsetIndex[:, 2],
                        ]
                    )
                    ** 2
                )

        return np.array(
            (E_Km, P_Km, ikpt - 1 * np.ones(len(E_Km))), dtype=float
        ).T

    # def spectral_weight(self, kpoints, nproc=None):
    #     '''
    #     Calculate the spectral weight for a list of kpoints in the primitive BZ.
    #     Here, we use "multiprocessing" package to parallel over the kpoints.
    #     '''
    #
    #     NKPTS = len(kpoints)
    #
    #     if nproc is None:
    #         nproc = multiprocessing.cpu_count()
    #
    #     pool = multiprocessing.Pool(processes=nproc)
    #
    #     results = []
    #     for ik in range(NKPTS):
    #         res = pool.apply_async(self.spectral_weight_k, (kpoints[ik],))
    #         results.append(res)
    #
    #     self.SW = np.array([res.get() for res in results], dtype=float)
    #
    #     pool.close()
    #     pool.join()
    #
    #     return self.SW

    def spectral_weight(self, kpoints):
        """
        Calculate the spectral weight for a list of kpoints in the primitive BZ.
        """

        NKPTS = len(kpoints)

        # self.SW = np.array([self.spectral_weight_k(kpoints[ik])
        #                     for ik in range(NKPTS)], dtype=float)
        sw = []
        for ispin in range(self.wfc._nspin):
            if self.wfc._nspin == 2:
                print("#" * 60)
                print("Spin component: %d" % ispin)
                print("#" * 60)
            sw.append(
                [
                    self.spectral_weight_k(kpoints[ik], whichspin=ispin + 1)
                    for ik in range(NKPTS)
                ]
            )

        self.SW = np.array(sw)

        # For noncollinear calculation, nspin = 1.
        # if self._lsoc:
        #     # self.SW = np.swapaxes(self.SW, 0, 1)
        #     self.SW = np.array([self.SW[0,:,:,0,:], self.SW[0,:,:,1,]])

        return self.SW

    def spectral_function(self, nedos=4000, sigma=0.02):
        """
        Generate the spectral function

            A(k_i, E) = \sum_m P_{Km}(k_i)\Delta(E - Em)

        Where the \Delta function can be approximated by Lorentzian or Gaussian
        function.
        """

        assert self.SW is not None, "Spectral weight must be calculated first!"

        # NS = 2 if self._lsoc else self.wfc._nspin
        # For noncollinear calculation, nspin = 1.
        NS = self.wfc._nspin
        # Number of kpoints
        nk = self.SW.shape[1]
        # spectral function
        SF = np.zeros((NS, nedos, nk), dtype=float)

        emin = self.SW[:, :, :, 0].min()
        emax = self.SW[:, :, :, 0].max()
        e0 = np.linspace(emin - 5 * sigma, emax + 5 * sigma, nedos)

        for ispin in range(NS):
            for ii in range(nk):
                E_Km = self.SW[ispin, ii, :, 0]
                P_Km = self.SW[ispin, ii, :, 1]

                SF[ispin, :, ii] = np.sum(
                    LorentzSmearing(
                        e0[:, np.newaxis], E_Km[np.newaxis, :], sigma=sigma
                    )
                    * P_Km[np.newaxis, :],
                    axis=1,
                )
        return e0, SF


############################################################

if __name__ == "__main__":
    M = [[3.0, 0.0, 0.0], [0.0, 3.0, 0.0], [0.0, 0.0, 1.0]]

    kpts = [
        [0.0, 0.5, 0.0],
        [0.0, 0.0, 0.0],
        [1.0 / 3, 1.0 / 3, 0.0],
        [0.0, 0.5, 0.0],
    ]

    kpath = make_kpath(kpts, nseg=30)

    K_in_sup = []
    for kk in kpath:
        kg, g = find_K_from_k(kk, M)
        K_in_sup.append(kg)
    reducedK = removeDuplicateKpoints(K_in_sup)
    save2VaspKPOINTS(reducedK)

    import os

    # from ase.io import read, write
    #
    # pos = read('POSCAR.p', format='vasp')
    # cell = pos.cell
    cell = [
        [3.1850, 0.0000000000000000, 0.0],
        [-1.5925, 2.7582909110534373, 0.0],
        [0.0000, 0.0000000000000000, 35.0],
    ]

    if os.path.isfile("spectral_function.npy"):
        sw = np.load("spectral_weight.npy")
        sf = np.load("spectral_function.npy")
        e0 = np.load("energy.npy")
    else:
        fwave = unfold(M=M, wavecar="WAVECAR")
        sw = fwave.spectral_weight(kpath)
        e0, sf = fwave.spectral_function(nedos=4000)
        np.save("spectral_weight.npy", sw)
        np.save("spectral_function.npy", sf)
        np.save("energy.npy", e0)

    EBS_scatter(
        kpath,
        cell,
        sw,
        nseg=30,
        eref=-4.01,
        ylim=(-3, 4),
        show=False,
        factor=5,
    )
    EBS_cmaps(
        kpath, cell, e0, sf, nseg=30, eref=-4.01, show=False, ylim=(-3, 4)
    )



from pymatgen.io.vasp.outputs import Chgcar
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.core.periodic_table import Element
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.operations import SymmOp
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from matplotlib.colors import Normalize
import numpy as np
from scipy.ndimage import gaussian_filter
import matplotlib.patheffects as pa
import copy as copy
from vaspvis.utils import make_supercell, group_layers
from ase.data.colors import jmol_colors
from sklearn.neighbors import radius_neighbors_graph
import os
import time


class STM:
    """
    This class contains all the methods for generating STM images with VASP
    """
    
    def __init__(
        self,
        folder,
        custom_color_dict = None,
    ):
        self.folder = folder
        self.preloaded_data = os.path.isfile(os.path.join(folder, 'parchg.npy'))
        self.poscar = Poscar.from_file(
            os.path.join(folder, 'POSCAR'),
            check_for_POTCAR=False,
            read_velocities=False
        )
        self.data, self.a_vals, self.b_vals, self.c_vals = self._load_parchg()
        [self.bottom_surface,
        self.bottom_ind,
        self.top_surface,
        self.top_ind] = self._get_surface_heights()
        self.X = None
        self.Y = None
        self.Z = None
        self.x_shift = None
        self.y_shift = None

    def _load_parchg(self):
        if self.preloaded_data:
            with open(os.path.join(self.folder, 'parchg.npy'), 'rb') as p:
                data = np.load(p)
        else:
            parchg = Chgcar.from_file(os.path.join(self.folder, 'PARCHG'))
            data = parchg.data['total']
            np.save(os.path.join(self.folder, 'parchg.npy'), data)

        a_vals = np.linspace(0,1,data.shape[0])
        b_vals = np.linspace(0,1,data.shape[1])
        c_vals = np.linspace(0,1,data.shape[2])

        return data, a_vals, b_vals, c_vals

    def _get_surface_heights(self):
        bottom_surface = self.poscar.structure.frac_coords[:,-1].min()
        top_surface = self.poscar.structure.frac_coords[:,-1].max()
        bottom_ind = np.argmin((self.c_vals - bottom_surface)**2)
        top_ind = np.argmin((self.c_vals - top_surface)**2)

        return bottom_surface, bottom_ind, top_surface, top_ind

    def _interp(self, x, x1, x2, y1, y2):
        return y1 + (((y2 - y1) / (x2 - x1)) * (x - x1)) 

    def _rotate_structure(self, structure, angle):
        copy_structure = copy.copy(structure)
        angle = angle * (np.pi / 180)
        operation = SymmOp.from_rotation_and_translation(
            rotation_matrix=np.array([
                [np.cos(angle), -np.sin(angle), 0],
                [np.sin(angle), np.cos(angle), 0],
                [0,0,1],
            ]),
            translation_vec=[0,0,0],
        )
        copy_structure.apply_operation(operation, fractional=False)

        return copy_structure

    def _get_constant_current_isosurface(self, current, sigma=6, top=True):
        slab_middle_ind = int((self.top_ind + self.bottom_ind) / 2)
        cell_middle_ind = int(self.data.shape[-1] / 2)
        shift = cell_middle_ind - slab_middle_ind
        init_shape = self.data.shape[:2]
        shifted_slab = np.roll(
            self.data,
            shift,
            axis=2,
        )
        c_vals = self.c_vals
        c_vals_extended = np.hstack([c_vals[:-1] - 1, c_vals, c_vals[1:] + 1])
        shifted_cvals = np.roll(
            c_vals_extended,
            shift,
        )
        shifted_cvals = shifted_cvals[len(c_vals)-1:(2*len(c_vals))-1]

        if top:
            shifted_slab = shifted_slab[:,:,self.top_ind+shift:]
            shifted_cvals = shifted_cvals[self.top_ind+shift:]
        else:
            shifted_slab = shifted_slab[:,:,:self.bottom_ind+shift]
            shifted_cvals = shifted_cvals[:self.bottom_ind+shift]


        if top:
            heights = np.zeros(shifted_slab.shape[:2])
            inds = np.zeros(shifted_slab.shape[:2], dtype=bool)
            for i in range(0, shifted_slab.shape[-1]-1)[::-1]:
                points = inds < (shifted_slab[:,:,i] > current)
                x1 = shifted_slab[points, i]
                x2 = shifted_slab[points, i+1]
                y1 = shifted_cvals[i]
                y2 = shifted_cvals[i+1]
                heights[points] = self._interp(
                    x=current,
                    x1=x1,
                    x2=x2,
                    y1=y1,
                    y2=y2,
                )
                inds[points] = True

            heights[heights <= self.top_surface] = heights[heights > self.top_surface].min()

        else:
            heights = np.zeros(shifted_slab.shape[:2])
            inds = np.zeros(shifted_slab.shape[:2], dtype=bool)
            for i in range(1, shifted_slab.shape[-1]):
                points = inds < (shifted_slab[:,:,i] > current)
                x1 = shifted_slab[points, i-1]
                x2 = shifted_slab[points, i]
                y1 = shifted_cvals[i-1]
                y2 = shifted_cvals[i]
                heights[points] = self._interp(
                    x=current,
                    x1=x1,
                    x2=x2,
                    y1=y1,
                    y2=y2,
                )
                inds[points] = True

            heights[heights >= self.top_surface] = heights[heights < self.top_surface].min()

        return heights

    def _generate_supercell(self, x, y, Z, scaling_matrix=[8,8]):
        x = np.concatenate([x + i for i in range(scaling_matrix[0])])
        y = np.concatenate([y + i for i in range(scaling_matrix[1])])
        Z = np.hstack([
            np.vstack([Z for _ in range(scaling_matrix[0])]) for _ in range(scaling_matrix[1])
        ])
        return x, y, Z

    def _get_intercept(self, midpoint, vector):
        if vector[0] == 0:
            intersect = [0, midpoint[1]]
        else:
            slope = vector[1] / vector[0]
            f = ((slope * midpoint[1]) + midpoint[0])/ ((slope**2) + 1)
            intersect = [f, slope * f]

        return intersect

    def _get_ratio(self, a, b):
        a_norm = np.linalg.norm(a)
        b_norm = np.linalg.norm(b)
        ratio_array = np.ones(2)
        min_ind = np.argmin([a_norm, b_norm])
        ratio = np.max([a_norm, b_norm]) / np.min([a_norm, b_norm])
        ratio_array[min_ind] = ratio
        
        return ratio_array

    def _get_square(self, a, b):
        midpoint = 0.5 * (a + b)
        a_inter = self._get_intercept(midpoint, a)
        b_inter = self._get_intercept(midpoint, b)
        a_len = np.linalg.norm(a_inter - midpoint)
        b_len = np.linalg.norm(b_inter - midpoint)

        r = np.min([a_len, b_len])

        box_length = (2 * r) / np.sqrt(2)

        return box_length, midpoint

    def _get_scaling_matrix(self, a, b, scan_size=40):
        final_box_length = 0
        final_midpoint = np.zeros(2)
        ratio = self._get_ratio(a, b)
        scaling_matrix = np.ones(2)
        i = 1
        while final_box_length <= scan_size:
            i += 1
            a_new = a * int(i * ratio[0])
            b_new = b * int(i * ratio[1])
            box_length, midpoint = self._get_square(a_new, b_new)
            final_box_length = box_length
            final_midpoint = midpoint
            scaling_matrix[0] = int(i * ratio[0])
            scaling_matrix[1] = int(i * ratio[1])

        return scaling_matrix.astype(int), midpoint

    def _run_constant_current_scan(self, current, structure, top=True, scan_size=40):
        scaling_matrix, midpoint = self._get_scaling_matrix(
            a=structure.lattice.matrix[0,:2],
            b=structure.lattice.matrix[1,:2],
            scan_size=scan_size,
        )
        Z = self._get_constant_current_isosurface(current, top=top)
        if top:
            Z = np.abs(Z - self.top_surface)
        else:
            Z = np.abs(Z - self.bottom_surface)

        x, y, Z = self._generate_supercell(
            self.a_vals,
            self.b_vals,
            Z,
            scaling_matrix=scaling_matrix,
        )

        X, Y = np.meshgrid(x, y)

        conv_input = np.c_[np.ravel(X), np.ravel(Y), np.ravel(Z)]
        converted = structure.lattice.get_cartesian_coords(conv_input)
        X_conv, Y_conv, Z_conv = converted[:,0], converted[:,1], converted[:,2]
        X_conv = X_conv.reshape(X.shape)
        Y_conv = Y_conv.reshape(Y.shape)
        Z_conv = Z_conv.reshape(Z.shape)
        shifted_point = midpoint - (scan_size / 2)
        X_conv -= shifted_point[0]
        Y_conv -= shifted_point[1]

        return X_conv, Y_conv, Z_conv, midpoint, scaling_matrix

    def _plot_stm_general(
        self,
        ax,
        X,
        Y,
        Z,
        cmap,
    ):
        ax.pcolormesh(
            X,
            Y,
            Z.T,
            shading='gouraud',
            cmap=cmap,
            norm=Normalize(vmin=Z.min(), vmax=Z.max()),
            rasterized=True,
            antialiased=True,
        )

    def _plot_atoms_general(
        self,
        ax,
        atol,
        max_bond_length,
        atom_size,
        bond_line_width,
        scaling_matrix,
        midpoint,
        scan_size,
        legend,
        top,
        structure,
        atom_axis_bounds,
        atoms_box,
        legend_atom_size,
        custom_color_dict=None,
    ):
        supercell = make_supercell(
            structure,
            scaling_matrix=np.hstack([scaling_matrix, 1]),
        )
        inds, heights = group_layers(supercell, atol=atol)

        if top:
            surface_inds = inds[-1]
        else:
            surface_inds = inds[0]

        surface_atom_coords = supercell.cart_coords[surface_inds]
        surface_atom_symbols = np.array(supercell.species, dtype='str')[surface_inds]
        surface_atom_species = np.zeros(surface_atom_symbols.shape, dtype=int)
        surface_atom_sizes = np.zeros(surface_atom_symbols.shape, dtype=float)
        unique_species = np.unique(surface_atom_symbols)
        unique_elements = [Element(i) for i in unique_species]
        unique_zs = [Element(i).Z for i in unique_species]

        for i, z in enumerate(unique_elements):
            surface_atom_species[np.isin(surface_atom_symbols, unique_species[i])] = z.Z
            surface_atom_sizes[np.isin(surface_atom_symbols, unique_species[i])] = z.atomic_radius

        surface_atom_sizes /= surface_atom_sizes.max()

        if custom_color_dict is not None:
            for element in custom_color_dict:
                z = Element(element).Z
                jmol_colors[z] = np.array(custom_color_dict[element]) / 255

        colors = jmol_colors[surface_atom_species]

        shifted_point = midpoint - (scan_size / 2)
        surface_atom_coords[:,0] -= shifted_point[0]
        surface_atom_coords[:,1] -= shifted_point[1]

        neighbor_graph = radius_neighbors_graph(
            X=surface_atom_coords,
            radius=max_bond_length,
        ).toarray()

        bonds = []

        for i in range(neighbor_graph.shape[0]):
            for j in range(neighbor_graph.shape[0]):
                if neighbor_graph[i,j] > 0:
                    to_append = [
                        surface_atom_coords[i],
                        surface_atom_coords[j],
                        [np.nan, np.nan, np.nan],
                    ]
                    bonds.append(to_append)

        bonds = np.vstack(bonds)

        ax_atoms = ax.inset_axes(
            bounds=atom_axis_bounds,
        )
        ax_atoms.set_xlim(
            atom_axis_bounds[0] * scan_size,
            (atom_axis_bounds[0] + atom_axis_bounds[2]) * scan_size
        )
        ax_atoms.set_ylim(
            atom_axis_bounds[1] * scan_size,
            (atom_axis_bounds[1] + atom_axis_bounds[3]) * scan_size
        )
        ax_atoms.set_facecolor((0,0,0,0))

        ax_atoms.tick_params(
            bottom=False,
            left=False,
            labelbottom=False,
            labelleft=False,
        )

        if not atoms_box:
            ax_atoms.spines['left'].set_visible(False)
            ax_atoms.spines['right'].set_visible(False)
            ax_atoms.spines['top'].set_visible(False)
            ax_atoms.spines['bottom'].set_visible(False)

        ax_atoms.plot(
            bonds[:,0],
            bonds[:,1],
            color='lightgrey',
            linewidth=bond_line_width,
            zorder=5,
            path_effects=[pa.Stroke(linewidth=bond_line_width+2, foreground='black'), pa.Normal()],
        )
        ax_atoms.scatter(
            surface_atom_coords[:,0],
            surface_atom_coords[:,1],
            c=colors,
            ec='black',
            s=atom_size*surface_atom_sizes,
            zorder=10,
        )

        if legend:
            legend_lines = []
            legend_labels = []
            for name, color, element in zip(unique_species, jmol_colors[unique_zs], unique_elements):
                legend_lines.append(plt.scatter(
                    [-1],
                    [-1],
                    color=color,
                    s=legend_atom_size*element.atomic_radius,
                    ec='black',
                ))
                legend_labels.append(
                    f'{name}'
                )

            leg = ax.get_legend()

            if leg is None:
                handles = legend_lines
                labels = legend_labels
            else:
                handles = [l._legmarker for l in leg.legendHandles]
                labels = [text._text for text in leg.texts]
                handles.extend(legend_lines)
                labels.extend(legend_labels)

            l = ax.legend(
                handles,
                labels,
                ncol=1,
                loc='upper right',
                framealpha=1,
                fontsize=16,
            )
            l.set_zorder(200)
            frame = l.get_frame()
            frame.set_facecolor('white')
            frame.set_edgecolor('black')

    def add_scale_bar(
        self,
        ax,
        width,
        height,
        offset_x=1,
        offset_y=1,
        units='nm',
        color='white',
        border_color='black',
        border_width=0.75,
        fontsize=16,
    ):
        rect = Rectangle(
            xy=(offset_x, offset_y),
            width=width,
            height=height,
            facecolor=color,
            edgecolor=border_color,
            linewidth=border_width,
            zorder=100,
        )
        if units == 'nm':
            annotation = f'{np.round(width / 10, 2)} nm'
        elif units == 'A':
            annotation = f'${width} \\AA$'
        else:
            raise('Please select proper units, either "nm" (nanometers) or "A" (angstroms)')

        ax.annotate(
            annotation,
            xy=(offset_x + (width / 2), offset_y + (1.05 * height)),
            xycoords='data',
            ha='center',
            va='bottom',
            color=color,
            fontsize=fontsize,
            fontweight='bold',
            path_effects=[pa.Stroke(linewidth=border_width, foreground=border_color)],
            zorder=100,
        )
        ax.add_patch(rect)


    def plot_constant_current(
        self,
        ax,
        current,
        top=True,
        scan_size=40,
        atol=0.03,
        plot_atoms=False,
        atom_size=80,
        legend_atom_size=100,
        bond_line_width=2,
        max_bond_length=3.14,
        cmap='hot',
        sigma=4,
        legend=False,
        rotation=0,
        atom_axis_bounds=[0.5,0.0,0.5,0.5],
        atoms_box=False,
        custom_color_dict=None,
    ):
        if rotation != 0:
            structure = self._rotate_structure(self.poscar.structure, angle=rotation)
        else:
            structure = self.poscar.structure

        X, Y, Z, midpoint, scaling_matrix = self._run_constant_current_scan(
            current=current,
            structure=structure,
            top=top,
            scan_size=scan_size,
        )
        Z = gaussian_filter(Z, sigma=sigma)
        shifted_midpoint = midpoint - (scan_size / 2)

        ax.set_xlim(0, scan_size)
        ax.set_ylim(0, scan_size)

        self._plot_stm_general(
            ax=ax,
            X=X,
            Y=Y,
            Z=Z,
            cmap=cmap,
        )

        if plot_atoms:
            self._plot_atoms_general(
                ax=ax,
                atol=atol,
                max_bond_length=max_bond_length,
                atom_size=atom_size,
                bond_line_width=bond_line_width,
                scaling_matrix=scaling_matrix,
                midpoint=midpoint,
                scan_size=scan_size,
                legend=legend,
                top=top,
                structure=structure,
                atom_axis_bounds=atom_axis_bounds,
                atoms_box=atoms_box,
                legend_atom_size=legend_atom_size,
                custom_color_dict=custom_color_dict,
            )


if __name__ == "__main__":
    fig, ax = plt.subplots(figsize=(4,4), dpi=400)
    ax.tick_params(
        bottom=False,
        left=False,
        labelbottom=False,
        labelleft=False,
    )

    stm = STM(folder='../../vaspvis_data/InAs111A_stm/')
    color_dict = {'In': [215, 128, 187], 'As': [116, 208, 87]}
    stm.plot_constant_current(
        ax=ax,
        current=0.009,
        top=False,
        scan_size=40,
        plot_atoms=True,
        sigma=5,
        cmap='Greys_r',
        atol=0.03,
        legend=True,
        rotation=0,
        #  atom_size=20,
        #  bond_line_width=0.5,
        legend_atom_size=80,
        atom_axis_bounds=[0.5, 0.0, 0.5, 1],
        atoms_box=True,
        custom_color_dict=color_dict,
    )
    stm.add_scale_bar(
        ax=ax,
        width=5,
        height=0.5,
        units='A',
        offset_x=2,
        offset_y=2,
        fontsize=20,
    )
    fig.tight_layout(pad=0)
    fig.savefig('InAs111A_with_atoms.png')




from pymatgen.io.vasp.outputs import Chgcar
from pymatgen.io.vasp.inputs import Poscar
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator
from multiprocessing import Pool
import numpy as np
import time
import os

class Charge:
    """
    This class contains all the methods for performing charge transfer calculations
    """

    def __init__(
        self,
        interface_folder,
        substrate_folder,
        film_folder,
    ):
        self.interface_folder = interface_folder
        self.substrate_folder = substrate_folder
        self.film_folder = film_folder
        self.interface_poscar = Poscar.from_file(
            os.path.join(self.interface_folder, 'POSCAR'),
            check_for_POTCAR=False,
            read_velocities=False
        )
        self.film_poscar = Poscar.from_file(
            os.path.join(self.film_folder, 'POSCAR'),
            check_for_POTCAR=False,
            read_velocities=False
        )
        self.substrate_poscar = Poscar.from_file(
            os.path.join(self.substrate_folder, 'POSCAR'),
            check_for_POTCAR=False,
            read_velocities=False
        )
        pool = Pool(processes=3)
        [self.interface_data, self.substrate_data, self.film_data] = pool.map(
            self._load_data,
            (self.interface_folder, self.substrate_folder, self.film_folder),
        )
        pool.close()
        pool.join()

    def _load_data(self, folder):
        preloaded_data = os.path.isfile(os.path.join(folder, 'chgcar.npy'))
        if preloaded_data:
            with open(os.path.join(folder, 'chgcar.npy'), 'rb') as p:
                data = np.load(p)
        else:
            chgcar = Chgcar.from_file(os.path.join(folder, 'CHGCAR'))
            data = chgcar.data['total']
            np.save(os.path.join(folder, 'chgcar.npy'), data)

        a_vals = np.linspace(0,1,data.shape[0])
        b_vals = np.linspace(0,1,data.shape[1])
        c_vals = np.linspace(0,1,data.shape[2])

        return data, a_vals, b_vals, c_vals

    def _planar_average(self, data):
        return data.mean(axis=(0,1))

    def _locate_ind(self, data, coord):
        return np.abs(data - coord).argmin()

    #  def macroscopic_average(potential, periodicity, resolution):
        #  """Getting the macroscopic average of potential
        #  Args:
            #  potential : array containig the electrostaticpotential/charge density
            #  periodicity : real number; the period over which to average
            #  resolution : the grid resolution in the direction of averaging
        #  Returns:
            #  macro_average : array with the macroscopically averaged values"""
#
#
        #  macro_average = np.zeros(shape=(len(potential)))
        #  period_points = int((periodicity/resolution))
        #  # Period points must be even
        #  if period_points % 2 != 0:
            #  period_points = period_points + 1
#
        #  length = len(potential)
        #  for i in range(length):
            #  start = i - int(period_points / 2)
            #  end = i + int(period_points / 2)
            #  if start < 0:
                #  start = start + length
                #  macro_average[i] = macro_average[i] + sum(potential[0:end]) + sum(potential[start:length])
                #  macro_average[i] = macro_average[i] / period_points
            #  elif end >= length:
                #  end = end - length
                #  macro_average[i] = macro_average[i] + sum(potential[start:length]) + sum(potential[0:end])
                #  macro_average[i] = macro_average[i] / period_points
            #  else:
                #  macro_average[i] = macro_average[i] + sum(potential[start:end]) / period_points
#
        #  print("Average of the average = ", numpy.average(macro_average))
#
        #  return macro_average

    def plot_charge_transfer(
        self,
        ax,
        color='black',
        linestyle='-',
        linewidth=1,
        interface_line_color='red',
        interface_linestyle='-',
        interface_linewidth=0.75,
        method=0,
        plot_interface_line=True,
    ):
        int_data, _, _, int_c = self.interface_data
        sub_data, _, _, sub_c = self.substrate_data
        film_data, _, _, film_c = self.film_data

        if method == 0:
            int_data = self._planar_average(int_data)
            sub_data = self._planar_average(sub_data)
            film_data = self._planar_average(film_data)

        int_min = self.interface_poscar.structure.frac_coords[:,-1].min()
        int_max = self.interface_poscar.structure.frac_coords[:,-1].max()
        sub_max = self.substrate_poscar.structure.frac_coords[:,-1].max()
        film_min = self.film_poscar.structure.frac_coords[:,-1].min()
        mid_line = (sub_max + film_min) / 2
        shift = 0.5 - mid_line
        mid_ind = self._locate_ind(int_c, mid_line)
        roll_val = int(int_c.shape[0] / 2) - mid_ind

        matrix = self.interface_poscar.structure.lattice.matrix
        area = np.linalg.norm(np.cross(matrix[0], matrix[1]))
        volume = np.linalg.det(matrix)
        c_len = np.linalg.norm(matrix[2])

        data = np.roll(int_data - sub_data - film_data, roll_val) * (area / volume)

        ax.plot(
            np.linspace(-c_len/2, c_len/2, int_c.shape[0]),
            data,
            color=color,
            linewidth=linewidth,
            linestyle=linestyle,
        )
        ax.set_xlim(-c_len/2,c_len/2)
        #  ax.xaxis.set_major_locator(MultipleLocator(0.25))

        if plot_interface_line:
            ax.axvline(
                x=0,
                color=interface_line_color,
                linewidth=interface_linewidth,
                linestyle=interface_linestyle,
            )


if __name__ == "__main__":
    chg = Charge(
        interface_folder='../../vaspvis_data/chgxfer/int',
        substrate_folder='../../vaspvis_data/chgxfer/sub',
        film_folder='../../vaspvis_data/chgxfer/film',
    )
    fig, ax = plt.subplots(figsize=(4,3), dpi=400)
    ax.set_ylabel('Charge Transfer $(e/\\AA)$')
    ax.set_xlabel('z-distance $(\\AA)$')
    chg.plot_charge_transfer(ax=ax)
    fig.tight_layout()
    fig.savefig('test.png')

        



import argparse
import numpy as np
import os
import copy
import spglib
from numpy.linalg import norm, solve
from pymatgen.io.vasp.inputs import Poscar
from ase.io import write, read
from ase.utils import gcd, basestring
from ase.build import bulk
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer


def surface(lattice, indices, layers, vacuum=None, tol=1e-10):
    """Create surface from a given lattice and Miller indices.

    lattice: Atoms object or str
        Bulk lattice structure of alloy or pure metal.  Note that the
        unit-cell must be the conventional cell - not the primitive cell.
        One can also give the chemical symbol as a string, in which case the
        correct bulk lattice will be generated automatically.
    indices: sequence of three int
        Surface normal in Miller indices (h,k,l).
    layers: int
        Number of equivalent layers of the slab.
    vacuum: float
        Amount of vacuum added on both sides of the slab.

    """

    indices = np.asarray(indices)

    if indices.shape != (3,) or not indices.any() or indices.dtype != int:
        raise ValueError('%s is an invalid surface type' % indices)

    if isinstance(lattice, basestring):
        lattice = bulk(lattice, cubic=True)

    h, k, l = indices
    h0, k0, l0 = (indices == 0)

    if h0 and k0 or h0 and l0 or k0 and l0:  # if two indices are zero
        if not h0:
            c1, c2, c3 = [(0, 1, 0), (0, 0, 1), (1, 0, 0)]
        if not k0:
            c1, c2, c3 = [(0, 0, 1), (1, 0, 0), (0, 1, 0)]
        if not l0:
            c1, c2, c3 = [(1, 0, 0), (0, 1, 0), (0, 0, 1)]
    else:
        p, q = ext_gcd(k, l)
        a1, a2, a3 = lattice.cell

        # constants describing the dot product of basis c1 and c2:
        # dot(c1,c2) = k1+i*k2, i in Z
        k1 = np.dot(p * (k * a1 - h * a2) + q * (l * a1 - h * a3),
                    l * a2 - k * a3)
        k2 = np.dot(l * (k * a1 - h * a2) - k * (l * a1 - h * a3),
                    l * a2 - k * a3)

        if abs(k2) > tol:
            i = -int(round(k1 / k2))  # i corresponding to the optimal basis
            p, q = p + i * l, q - i * k

        a, b = ext_gcd(p * k + q * l, h)

        c1 = (p * k + q * l, -p * h, -q * h)
        c2 = np.array((0, l, -k)) // abs(gcd(l, k))
        c3 = (b, a * p, a * q)
    surf = build(lattice, np.array([c1, c2, c3]), layers, tol)
    if vacuum is not None:
        surf.center(vacuum=vacuum, axis=2)
    return surf


def build(lattice, basis, layers, tol):
    surf = lattice.copy()
    scaled = solve(basis.T, surf.get_scaled_positions().T).T
    scaled -= np.floor(scaled + tol)
    surf.set_scaled_positions(scaled)
    surf.set_cell(np.dot(basis, surf.cell), scale_atoms=True)
    surf *= (1, 1, layers)
    a1, a2, a3 = surf.cell
    surf.set_cell([a1, a2,
                   np.cross(a1, a2) * np.dot(a3, np.cross(a1, a2)) /
                   norm(np.cross(a1, a2))**2])
    surf.pbc = (True, True, False)

    # Move atoms into the unit cell:
    scaled = surf.get_scaled_positions()
    scaled[:, :2] %= 1
    surf.set_scaled_positions(scaled)

    surf.cell[2] = 0.0

    return surf


def ext_gcd(a, b):
    if b == 0:
        return 1, 0
    elif a % b == 0:
        return 0, 1
    else:
        x, y = ext_gcd(b, a % b)
        return y, x - y * (a // b)


def convert(bulk, slab, index, output, generate=True, print_M=True):
    if type(bulk) == str:
        primitiveCell = Structure.from_file(bulk)
    else:
        primitiveCell = bulk
    if type(slab) == str:
        refSlab = Structure.from_file(slab)
    else:
        refSlab = slab

    sa = SpacegroupAnalyzer(primitiveCell)
    conventionalCell = sa.get_conventional_standard_structure()
    conventionalCell.to(filename='POSCAR.conventional')
    bulk = read('POSCAR.conventional')
    os.remove('POSCAR.conventional')
    slab = surface(bulk, index, layers=2, vacuum=10)
    lattice, _, _ = spglib.standardize_cell(cell=(slab.get_cell(
    ), slab.get_scaled_positions(), slab.get_atomic_numbers()), no_idealize=True)
    lattice_params = np.sort(np.linalg.norm(lattice, axis=1))[:2]
    scales = np.round(np.array([refSlab.lattice.a, refSlab.lattice.b] / lattice_params), 2)
    newLattice = []
    oldLattice = refSlab.lattice
    for length, scale in zip([oldLattice.a, oldLattice.b], scales):
        for j in range(len(lattice)):
            if abs((np.linalg.norm(lattice[j]) * scale) - length) < 1e-1:
                newLattice.append(copy.copy(scale * lattice[j][:]))
                lattice[j] = [0, 0, 0]
                break
    for i in range(len(lattice)):
        norm = np.linalg.norm(lattice[i])
        if norm > 1e-1:
            newLattice.append(lattice[i] / norm * oldLattice.c)
            break
    newLattice = Lattice(np.array(newLattice))
    for x, y in zip(oldLattice.angles, newLattice.angles):
        assert abs(x-y) < 1e-2, "The converted lattice has incorrect angles: {} compared with reference slab: {}".format(
            " ".join(str(x) for x in newLattice.angles), " ".join(str(x) for x in oldLattice.angles))
    newSlab = Structure(newLattice, [], [])
    for atom in refSlab:
        newSlab.append(atom.specie, atom.frac_coords[:])

    if generate:
        Poscar(newSlab.get_sorted_structure()).write_file(output, direct=True)

    transformMat = newSlab.lattice.matrix.dot(
        np.linalg.inv(primitiveCell.lattice.matrix))
    transformMat = transformMat.round().astype(int)
    if print_M:
        print('-------------------------------------------')
        print('Your Transformtaion Matrix is:')
        print(' ')
        print(transformMat)
        print('-------------------------------------------')

    return transformMat

if __name__ == "__main__":
    bulk = './POSCAR_bulk'
    slab = './POSCAR_10_100_22'
    index = [1,0,0]
    #  bulk = './POSCAR_HgTe_ept'
    #  slab = './POSCAR_18'
    #  index = [1,0,0]
    output = 'POSCAR_conv2'
    M = convert(bulk, slab, index, output)
